"""File-related utility functions.

Implements: `load_file`, `dump_file`, `load_json_file`,
`dump_json_file`, `get_parent_dir_path`, `get_dir_checksum`,
`get_file_checksum`, `load_raw_proffast_output`"""

from __future__ import annotations
from typing import Any, Optional
import hashlib
import json
import os
import polars as pl
import tum_esm_utils


def load_file(path: str) -> str:
    with open(path, "r") as f:
        return f.read()


def dump_file(path: str, content: str) -> None:
    with open(path, "w") as f:
        f.write(content)


def load_json_file(path: str) -> Any:
    with open(path, "r") as f:
        return json.load(f)


def dump_json_file(path: str, content: Any, indent: Optional[int] = 4) -> None:
    with open(path, "w") as f:
        json.dump(content, f, indent=indent)


def get_parent_dir_path(script_path: str, current_depth: int = 1) -> str:
    """Get the absolute path of a parent directory based on the
    current script path. Simply pass the `__file__` variable of
    the current script to this function. Depth of 1 will return
    the direct parent directory of the current script."""
    assert current_depth > 0, "depth must be greater than 0"
    output = os.path.dirname(os.path.abspath(script_path))
    for _ in range(current_depth - 1):
        output = os.path.dirname(output)
    return output


def get_dir_checksum(path: str) -> str:
    """Get the checksum of a directory using md5deep."""

    assert os.path.isdir(path), f"{path} is not a directory"
    return tum_esm_utils.shell.run_shell_command(
        f"md5deep -r -b {path} | sort -u | md5sum"
    )


def get_file_checksum(path: str) -> str:
    """Get the checksum of a file using MD5 from `haslib`.

    Significantly faster than `get_dir_checksum` since it does
    not spawn a new process."""

    assert os.path.isfile(path), f"{path} is not a file"
    with open(path, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


def load_raw_proffast_output(
    path: str,
    selected_columns: list[str] = [
        "gnd_p",
        "gnd_t",
        "app_sza",
        "azimuth",
        "xh2o",
        "xair",
        "xco2",
        "xch4",
        "xco",
        "xch4_s5p",
    ],
) -> pl.DataFrame:
    """
    Returns a raw proffast output file as a dataframe.

    you can pass `selected_columns` to only keep some columns - the
    `utc` column will always be included. Example:

    ```
    utc                     gnd_p    gnd_t    app_sza   ...
    2021-10-20 07:00:23     950.91   289.05   78.45     ...
    2021-10-20 07:00:38     950.91   289.05   78.42     ...
    2021-10-20 07:01:24     950.91   289.05   78.31     ...
    ...                     ...      ...      ...       ...
    [1204 rows x 8 columns]
    ```
    """

    assert os.path.isfile(path), f"{path} is not a file"

    data_column_names = {
        "gnd_p": " gndP",
        "gnd_t": " gndT",
        "app_sza": " appSZA",
        "azimuth": " azimuth",
        "xh2o": " XH2O",
        "xair": " XAIR",
        "xco2": " XCO2",
        "xch4": " XCH4",
        "xco": " XCO",
        "xch4_s5p": " XCH4_S5P",
    }

    assert len(set(selected_columns)) == len(
        selected_columns
    ), "selected_columns cannot contain duplicate items"
    assert set(selected_columns).issubset(set(data_column_names.keys())), (
        f"selected_columns contains invalid items, only the " +
        f"following are allowed: {data_column_names.keys()}"
    )

    df = pl.read_csv(
        path,
        columns=[
            "UTC",
            *list(data_column_names.values()),
        ],
        new_columns=[
            "utc",
            *list(data_column_names.keys()),
        ],
        dtypes={
            "utc": pl.Datetime,
            **{t: pl.Float32
               for t in data_column_names.keys()},
        },
    )

    # only keep the selected columns
    return df.select(["utc", *selected_columns])



"""Implements validator functions for use with pydantic models.

Implements: `validate_bool`, `validate_float`, `validate_int`,
`validate_str`, `validate_list`"""

from __future__ import annotations
from typing import Any, Callable, Optional, TypeVar
import os
import re
import datetime
import warnings


# duplicate method because lazydocs complains when using relative imports
def _is_date_string(date_string: str) -> bool:
    """Returns true if string is in `YYYYMMDD` format and date exists"""
    try:
        datetime.datetime.strptime(date_string, "%Y%m%d")
        return True
    except (AssertionError, ValueError):
        return False


# duplicate method because lazydocs complains when using relative imports
def _is_datetime_string(datetime_string: str) -> bool:
    """Returns true if string is in `YYYYMMDD HH:mm:ss` format and date exists"""

    warnings.warn(
        "Use `is_rfc3339_datetime_string` instead", DeprecationWarning
    )
    try:
        datetime.datetime.strptime(datetime_string, "%Y%m%d %H:%M:%S")
        return True
    except (AssertionError, ValueError):
        return False


# duplicate method because lazydocs complains when using relative imports
def _is_rfc3339_datetime_string(datetime_string: str) -> bool:
    """Returns true if string is in `YYYY-MM-DDTHH:mm:ssZ` (RFC3339)
    format and date exists. Caution: The appendix of `+00:00` is required for UTC!"""
    try:
        assert re.match(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\+|\-)\d{2}:\d{2}$",
            datetime_string,
        )
        datetime.datetime.fromisoformat(datetime_string)
        return True
    except (ValueError, AssertionError):
        return False


def validate_bool() -> Callable[[Any, bool], bool]:
    def f(cls: Any, v: Any) -> bool:
        if not isinstance(v, bool):
            raise ValueError(f'"{v}" is not a boolean')
        return v

    return f


def validate_float(
    nullable: bool = False,
    minimum: Optional[float] = None,
    maximum: Optional[float] = None,
) -> Callable[[Any, float], float]:
    def f(cls: Any, v: float) -> float:
        if v is None:
            if nullable:
                return v
            else:
                raise ValueError(f"value cannot be None")
        if not isinstance(v, (int, float)):
            raise ValueError(f'"{v}" is not a float')
        if minimum is not None and v < minimum:
            raise ValueError(f'"{v}" is smaller than {minimum}')
        if maximum is not None and v > maximum:
            raise ValueError(f'"{v}" is larger than {maximum}')
        return v

    return f


def validate_int(
    nullable: bool = False,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    allowed: Optional[list[int]] = None,
    forbidden: Optional[list[int]] = None,
) -> Callable[[Any, Optional[int]], Optional[int]]:
    def f(cls: Any, v: Optional[int]) -> Optional[int]:
        if v is None:
            if nullable:
                return v
            else:
                raise ValueError(f"value cannot be None")
        if not isinstance(v, int):
            raise ValueError(f'"{v}" is not an integer')
        if minimum is not None and v < minimum:
            raise ValueError(f'"{v}" is smaller than {minimum}')
        if maximum is not None and v > maximum:
            raise ValueError(f'"{v}" is larger than {maximum}')
        if allowed is not None and v not in allowed:
            raise ValueError(f'"{v}" is not allowed (not one of {allowed})')
        if forbidden is not None and v in forbidden:
            raise ValueError(f'"{v}" is forbidden)')
        return v

    return f


def validate_str(
    nullable: bool = False,
    min_len: Optional[float] = None,
    max_len: Optional[float] = None,
    regex: Optional[str] = None,
    is_numeric: bool = False,
    is_directory: bool = False,
    is_file: bool = False,
    is_date_string: bool = False,
    is_datetime_string: bool = False,
    is_rfc3339_datetime_string: bool = False,
    allowed: Optional[list[str]] = None,
    forbidden: Optional[list[str]] = None,
) -> Callable[[Any, str], str]:
    def f(cls: Any, v: str) -> str:
        if v is None:
            if nullable:
                return v
            else:
                raise ValueError(f"value cannot be None")
        if not isinstance(v, str):
            raise ValueError(f'"{v}" is not a string')

        # length and regex validation
        if min_len is not None and len(v) < min_len:
            raise ValueError(f'"{v}" has less than {min_len} characters')
        if max_len is not None and len(v) > max_len:
            raise ValueError(f'"{v}" has more than {max_len} characters')
        if regex is not None and re.compile(regex).match(v) is None:
            raise ValueError(f'"{v}" does not match the regex "{regex}"')
        if is_numeric and not v.isnumeric():
            raise ValueError(f'"{v}" is not numeric')

        # directory and file validation
        if is_directory and not os.path.isdir(v):
            raise ValueError(f'"{v}" is not a directory')
        if is_file and not os.path.isfile(v):
            raise ValueError(f'"{v}" is not a file')

        # date and datetime string validation
        if is_date_string and not _is_date_string(v):
            raise ValueError(f'"{v}" is not a date string ("YYYYMMDD")')
        if is_datetime_string and not _is_datetime_string(v):
            raise ValueError(
                f'"{v}" is not a datetime string ("YYYYMMDD HH:mm:ss")'
            )
        if is_rfc3339_datetime_string and not _is_rfc3339_datetime_string(v):
            raise ValueError(
                f'"{v}" is not a datetime string ("YYYY-MM-DDTHH:mm:ssZ")'
            )

        # allowed and forbidden values
        if allowed is not None and v not in allowed:
            raise ValueError(f'"{v}" is not allowed (not one of {allowed})')
        if forbidden is not None and v in forbidden:
            raise ValueError(f'"{v}" is forbidden)')
        return v

    return f


T = TypeVar("T")


def validate_list(
    min_len: Optional[float] = None,
    max_len: Optional[float] = None,
    allowed: Optional[list[T]] = None,
    required: Optional[list[T]] = None,
    forbidden: Optional[list[T]] = None,
) -> Callable[[Any, list[T]], list[T]]:
    def f(cls: Any, v: list[T]) -> list[T]:
        if not isinstance(v, list):
            raise ValueError(f'"{v}" is not a list')
        if min_len is not None and len(v) < min_len:
            raise ValueError(f'"{v}" has less than {min_len} elements')
        if max_len is not None and len(v) > max_len:
            raise ValueError(f'"{v}" has more than {max_len} elements')
        if allowed is not None:
            for item in v:
                if item not in allowed:
                    raise ValueError(f'"{item}" is not allowed')
        if required is not None:
            for item in required:
                if item not in v:
                    raise ValueError(f'"{v}" has to include "{item}"')
        if forbidden is not None:
            for item in forbidden:
                if item in v:
                    raise ValueError(f'"{v}" should not include "{item}"')
        return v

    return f



"""Implements custom logging functionality, because the
standard logging module is hard to configure for special
cases.

Implements: `Logger`"""

from __future__ import annotations
from typing import Literal, Optional
import os
import sys
import traceback
import datetime
import filelock


# duplicate method because lazydocs complains when using relative imports
def _pad_string(
    text: str,
    min_width: int,
    pad_position: Literal["left", "right"] = "left",
    fill_char: Literal["0", " ", "-", "_"] = " ",
) -> str:
    if len(text) >= min_width:
        return text
    else:
        pad = fill_char * (min_width - len(text))
        return (pad + text) if (pad_position == "left") else (text + pad)


# duplicate method because lazydocs complains when using relative imports
def _get_utc_offset() -> float:
    """Returns the UTC offset of the system"""
    return round((datetime.datetime.now() -
                  datetime.datetime.utcnow()).total_seconds() / 3600, 1)


# The logging module behaved very weird with the setup we have
# therefore I am just formatting and appending the log lines
# manually. Doesn't really make a performance difference.


def _log_line_has_date(log_line: str) -> bool:
    """Checks whether a given log line (string) starts with a valid date.

    This is not true for exception tracebacks. This log line time is used
    to determine which file to archive logs lines into"""
    try:
        datetime.datetime.strptime(log_line[: 10], "%Y-%m-%d")
        return True
    except:
        return False


class Logger:
    last_archive_time = datetime.datetime.now()

    def __init__(
        self,
        origin: str,
        logfile_directory: str,
        write_to_files: bool = True,
        print_to_console: bool = False,
    ) -> None:
        self.origin: str = origin

        self.filelock = filelock.FileLock(
            os.path.join(logfile_directory, "logging.lock"), timeout=30
        )
        self.current_logfile_path = os.path.join(
            logfile_directory, "current-logs.log"
        )
        self.log_archive_path = os.path.join(logfile_directory, "archive")

        self.write_to_files = write_to_files
        self.print_to_console = print_to_console

    def horizontal_line(
        self, fill_char: Literal["-", "=", ".", "_"] = "="
    ) -> None:
        """writes a horizonal line wiht `-`/`=`/... characters"""
        self._write_log_line("INFO", fill_char * 46)

    def debug(
        self,
        message: str,
        details: Optional[str] = None,
    ) -> None:
        """writes a debug log line"""
        self._write_log_line("DEBUG", message, details=details)

    def info(
        self,
        message: str,
        details: Optional[str] = None,
    ) -> None:
        """writes an info log line"""
        self._write_log_line("INFO", message, details=details)

    def warning(
        self,
        message: str,
        details: Optional[str] = None,
    ) -> None:
        """writes a warning log line"""
        self._write_log_line("WARNING", message, details=details)

    def error(
        self,
        message: str,
        details: Optional[str] = None,
    ) -> None:
        """writes an error log line, sends the message via
        MQTT when config is passed (required for revision number)
        """
        self._write_log_line("ERROR", message, details=details)

    def exception(
        self,
        label: Optional[str] = None,
        details: Optional[str] = None,
    ) -> None:
        """logs the traceback of an exception; output will be
        formatted like this:

        ```
        (label, )ZeroDivisionError: division by zero
        --- details: -----------------
        ...
        --- traceback: ---------------
        ...
        ------------------------------
        ```
        """
        exc_type, exc, exc_traceback = sys.exc_info()
        exception_name = traceback.format_exception_only(exc_type,
                                                         exc)[0].strip()
        exception_traceback = "\n".join(
            traceback.format_exception(exc_type, exc, exc_traceback)
        ).strip()

        subject_string = (
            exception_name if label is None else f"{label}, {exception_name}"
        )

        self._write_log_line(
            "EXCEPTION",
            subject_string,
            details=details,
            traceback=exception_traceback,
        )

    def _write_log_line(
        self,
        level: str,
        message: str,
        **kwargs: Optional[str],
    ) -> None:
        """Formats the log line string and writes it to `logs/current-logs.log`.
        All keyword arguments will be added to the log message with horizontal
        dividers. With `kwargs={"a": "somevalue", "b": None}` the log message will
        look like this:

        ```
        time UTC±X - origin - level - message
        --- a: ---------------------------------
        somevalue
        ----------------------------------------
        ```
        """
        now = datetime.datetime.now()
        utc_offset = _get_utc_offset()
        if utc_offset % 1 == 0:
            utc_offset = round(utc_offset)

        additional_log_items = [(k, v)
                                for k, v in kwargs.items() if v is not None]
        if len(additional_log_items) > 0:
            message += "\n"
            for label, value in additional_log_items:
                message += _pad_string(
                    f"--- {label}: ---",
                    min_width=40,
                    pad_position="right",
                    fill_char="-",
                )
                message += "\n" + value + "\n"
            message += "-" * 40

        log_string = (
            f"{str(now)[:-3]} UTC{'-' if utc_offset < 0 else '+'}{abs(utc_offset)} "
            +
            f"- {_pad_string(self.origin, min_width=23, pad_position='right')} "
            + f"- {_pad_string(level, min_width=13, pad_position='right')} " +
            f"- {message}"
        )
        if self.print_to_console:
            print(log_string)
        else:
            with self.filelock:
                with open(self.current_logfile_path, "a") as f1:
                    f1.write(log_string + "\n")

                # Archive lines older than 60 minutes, every 10 minutes
                if (now - Logger.last_archive_time).total_seconds() > 600:
                    self._archive()

    def _archive(self) -> None:
        """moves old log lines in "logs/current-logs.log" into an
        archive file "logs/archive/YYYYMMDD.log". log lines from
        the last hour will remain"""

        with open(self.current_logfile_path) as f:
            log_lines_in_file = f.readlines()
        if len(log_lines_in_file) == 0:
            return

        lines_to_be_kept, lines_to_be_archived = [], []
        latest_time = str(datetime.datetime.now() - datetime.timedelta(hours=1))
        line_time = log_lines_in_file[0][: 26]
        for index, line in enumerate(log_lines_in_file):
            if _log_line_has_date(line):
                line_time = line[: 26]
            if line_time > latest_time:
                lines_to_be_archived = log_lines_in_file[: index]
                lines_to_be_kept = log_lines_in_file[index :]
                break

        with open(self.current_logfile_path, "w") as f:
            f.writelines(lines_to_be_kept)

        if len(lines_to_be_archived) == 0:
            return

        archive_log_date_groups: dict[str, list[str]] = {}
        line_date = lines_to_be_archived[0][: 10].replace("-", "")
        for line in lines_to_be_archived:
            if _log_line_has_date(line):
                line_date = line[: 10].replace("-", "")
            if line_date not in archive_log_date_groups.keys():
                archive_log_date_groups[line_date] = []
            archive_log_date_groups[line_date].append(line)

        for date in archive_log_date_groups.keys():
            filename = os.path.join(self.log_archive_path, f"{date}.log")
            with open(filename, "a") as f:
                f.writelines(archive_log_date_groups[date] + [""])

        Logger.last_archive_time = datetime.datetime.now()



"""Context managers for common tasks.

Implements: `ensure_section_duration`"""

from typing import Generator
import contextlib
import time


@contextlib.contextmanager
def ensure_section_duration(duration: float) -> Generator[None, None, None]:
    """Make sure that the duration of the section is at least the given duration.

    Usage example - do one measurement every 6 seconds:

    ```python
    with ensure_section_duration(6):
        do_measurement()
    ```
    """

    start_time = time.time()
    yield
    remaining_loop_time = start_time + duration - time.time()
    if remaining_loop_time > 0:
        time.sleep(remaining_loop_time)



"""Common system status related functions.

Implements: `get_cpu_usage`, `get_memory_usage`, `get_disk_space`,
`get_system_battery`, `get_last_boot_time`, `get_utc_offset`"""

from __future__ import annotations
from typing import Any
import psutil
import datetime


def get_cpu_usage() -> list[float]:
    """Returns cpu_percent for all cores as `list[cpu1%, cpu2%,...]`"""
    return psutil.cpu_percent(interval=1, percpu=True)


def get_memory_usage() -> float:
    """Returns the memory usage in %"""
    p = psutil.virtual_memory().percent
    assert isinstance(p, float)
    return p


def get_disk_space() -> float:
    """Returns disk space used in % as float"""
    return psutil.disk_usage("/").percent


def get_system_battery() -> int:
    """Returns system battery in percent in percent.
    Returns 100 if device has no battery."""

    # FIXME: In the next breaking release, return None if device has no battery

    battery_state: Any = psutil.sensors_battery()  # type: ignore
    try:
        assert battery_state is not None
        percent = battery_state.percent
        assert isinstance(percent, int)
        assert 1 <= percent <= 100
        return percent
    except AssertionError:
        return 100


def get_last_boot_time() -> str:
    """Returns last OS boot time."""

    # FIXME: In the next breaking release, convert this to a return a datetime

    return datetime.datetime.fromtimestamp(psutil.boot_time()
                                          ).strftime("%Y-%m-%d %H:%M:%S")


def get_utc_offset() -> float:
    """Returns the UTC offset of the system.

    ```python
    x = get_utc_offset()
    local time == utc time + x
    ```
    """
    return round((datetime.datetime.now() -
                  datetime.datetime.utcnow()).total_seconds() / 3600, 1)



"""Functions for interacting with interferograms.

Implements: `detect_corrupt_ifgs`"""

from __future__ import annotations
from typing import Literal
import os
import re
import subprocess
import filelock
import tum_esm_utils

_PARSER_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "ifg_parser"
)


def _compile_fortran_code(
    silent: bool = True,
    fortran_compiler: Literal["gfortran", "gfortran-9"] = "gfortran",
) -> None:
    if not os.path.isfile(os.path.join(_PARSER_DIR, "ifg_parser")):
        if not silent:
            print("compiling fortran code")

        command = f"{fortran_compiler} -nocpp -O3 -o ./ifg_parser glob_prepro4.F90 glob_OPUSparms.F90 ifg_parser.F90"
        p = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=_PARSER_DIR,
            env=os.environ.copy(),
        )
        if p.returncode != 0:
            stdout = p.stdout.decode("utf-8", errors="replace").strip()
            stderr = p.stderr.decode("utf-8", errors="replace").strip()
            raise Exception(
                f"command '{command}' failed with exit code {p.returncode}, " +
                f"stderr: {stderr}, stout:{stdout}",
            )


def _write_input_file(random_id: str, ifgs: list[str]) -> None:
    with open(f"{_PARSER_DIR}/ifg_parser.template.inp", "r") as f:
        template_content = f.read()
    with open(f"{_PARSER_DIR}/ifg_parser.inp.{random_id}", "w") as f:
        f.write(template_content.replace("%IFG_LIST%", "\n".join(ifgs)))


def detect_corrupt_ifgs(
    ifg_directory: str,
    silent: bool = True,
    fortran_compiler: Literal["gfortran", "gfortran-9"] = "gfortran",
) -> dict[str, list[str]]:
    """Returns dict[filename, list[error_messages]] for all
    corrupt interferograms in the given directory.

    It will compile the fortran code using a given compiler
    to perform this task. The fortran code is derived from
    the preprocess source code of Proffast 2
    (https://www.imk-asf.kit.edu/english/3225.php). We use
    it because the retrieval using Proffast 2 will fail if
    there are corrupt interferograms in the input."""

    # compiling fortran code
    _compile_fortran_code(silent=silent, fortran_compiler=fortran_compiler)

    # generate input file
    ifgs = [f"{ifg_directory}/{x}" for x in os.listdir(ifg_directory)]
    ifgs = list(sorted(list(filter(os.path.isfile, ifgs))))

    # run the parser
    results: dict[str, list[str]] = {}
    stdout: str = ""
    while True:
        with filelock.FileLock(f"{_PARSER_DIR}/ifg_parser.inp.lock"):
            random_id = tum_esm_utils.text.get_random_string(
                10,
                forbidden=[
                    f.replace("ifg_parser.inp.", "")
                    for f in os.listdir(_PARSER_DIR)
                    if f.startswith("ifg_parser.inp.")
                ],
            )
            _write_input_file(random_id, ifgs)
        process = subprocess.run(
            ["./ifg_parser", f"ifg_parser.inp.{random_id}"],
            cwd=_PARSER_DIR,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        os.remove(f"{_PARSER_DIR}/ifg_parser.inp.{random_id}")
        stdout = process.stdout.decode()
        stderr = process.stderr.decode()
        if process.returncode == 0:
            break
        else:
            # find the filename that caused the error
            failing_filenames = list(
                filter(lambda f: f in ifgs, stderr.split("'"))
            )
            assert (
                "At line 837 of file ifg_parser.F90" in stderr
            ), f"Unknown error behavior: {stderr}"
            assert len(
                failing_filenames
            ) == 1, "invalid filename not found in stderr"
            failing_filepath = failing_filenames[0]

            # remove the error-causing file from the ifg list and try again
            ifgs.remove(failing_filepath)
            failing_filename = failing_filepath.split("/")[-1]
            results[failing_filename] = ["file not processable"]
            if not silent:
                print(f'error with file "{failing_filename}", running again')

    with open(os.path.join(_PARSER_DIR, "output.txt"), "w") as f:
        f.write(stdout)

    file_parsing_block = stdout.split("Done!")[-1]
    file_parsing_lines = file_parsing_block.split("Read OPUS parms:")[1 :]

    # get results from output stream
    for line in file_parsing_lines:
        filename = line[12 :].split("\n")[0].split("/")[-1].replace(")", "")
        parser_output = line[12 :].replace("\n", " ")
        parser_messages = re.findall(
            'charfilter "[^"]+" is missing', parser_output
        )
        if len(parser_messages) > 0:
            results[filename] = [
                ("charfilter '" + x.split('"')[1] + "' is missing")
                for x in parser_messages
            ]

    return results
